open Fof
open Term
open Utils

let rec fold_right1 f = function
    x :: [] -> x
  | x :: xs -> f x (fold_right1 f xs)
  | [] -> failwith "fold_right1"

let eq_refl eqn = Forall (1, Atom (eqn, [V 1; V 1]))
let eq_sym eqn =
  Forall (1, Forall (2, Impl (Atom (eqn, [V 1; V 2]), Atom (eqn, [V 2; V 1]))))
let eq_trans eqn = Forall (1, Forall (2, Forall (3, Impl (Conj (Atom (eqn, [V 1; V 2]), Atom (eqn, [V 2; V 3])), Atom (eqn, [V 1; V 3])))))

let subst_axiom2 eqn isfun (p, arity) =
  let even = List.init arity (fun i -> 2 * i) in
  let odd  = List.map (fun x -> x + 1) even in
  let ovar = List.map (fun x -> V x) odd
  and evar = List.map (fun x -> V x) even in
  let eqs = List.map2 (fun x y -> Atom (eqn, [x; y])) evar ovar in
  let eql = fold_right1 conj eqs in
  let eqr = if isfun then Impl (eql, Atom (eqn, [A (p, evar); A (p, ovar)]))
            else Impl (Conj (eql, Atom (p, evar)), Atom (p, ovar)) in
  List.fold_right forall (even @ odd) eqr

let add_antecedent a = function
    Impl (b, c) -> Impl (Conj (a, b), c)
  | c -> Impl (a, c)

let add_conj acc = function
    [] -> acc
  | l -> Conj (acc, fold_right1 conj l)

let equal_axioms eqn form =
  let eqp = eqn, 2
  and (pred, fnct) = collect_predfunc form in
  if not (List.mem eqp pred) then form
  else
    let pred = List.filter ((<>) eqp) pred in
    let basic = Conj (eq_refl eqn, Conj (eq_sym eqn, eq_trans eqn)) in
    let fnax = List.map (subst_axiom2 eqn  true) fnct
    and prax = List.map (subst_axiom2 eqn false) pred in
    let all = add_conj (add_conj basic prax) fnax in
    add_antecedent all form
