type term = V of int
          | A of int * term list;;

exception Unify;;

exception Unchanged;;

let qpartition p m =
  let rec qpartition l =
    if l == m then raise Unchanged else
    match l with
      [] -> raise Unchanged
    | (h::t) -> if p h then
        try let yes,no = qpartition t in h::yes,no
        with Unchanged -> [h],t
      else
        let yes,no = qpartition t in yes,h::no in
    function l -> try qpartition l
                  with Unchanged -> [],l;;

let rec rev_assoc a l =
  match l with
    (x,y)::t -> if Pervasives.compare y a = 0 then x else rev_assoc a t
  | [] -> raise Not_found;;

let rec istriv env x t =
  match t with
    V y -> y = x or
              (try let t' = rev_assoc y env in istriv env x t'
               with Not_found -> false)
  | A(f,args) -> List.exists (istriv env x) args & raise Unify;;

let rec itlist2 f l1 l2 b =
  match (l1,l2) with
    ([],[]) -> b
  | (h1::t1,h2::t2) -> f h1 h2 (itlist2 f t1 t2 b)
  | _ -> failwith "itlist2";;

let offinc = 10000;;

let rec rev_assocd a l d =
  match l with
    [] -> d
  | (x,y)::t -> if Pervasives.compare y a = 0 then x else rev_assocd a t d;;

let rec qmap f l =
  match l with
    h::t -> let h' = f h and t' = qmap f t in
            if h' == h & t' == t then l else h'::t'
  | _ -> l;;

let rec fol_subst_bump offset theta tm =
    match tm with
      V v -> (*if v < offinc then*)
                 let v' = v + offset in
                 rev_assocd v' theta (V(v'))
               (*else
                 rev_assocd v theta tm*)
    | A(f,args) ->
          let args' = qmap (fol_subst_bump offset theta) args in
          if args' == args then tm else A(f,args');;

let fol_inst_bump offset theta ((p,args) as at) =
    let args' = qmap (fol_subst_bump offset theta) args in
    if args' == args then at else p,args';;

let rec fol_unify offset tm1 tm2 sofar =
    match tm1,tm2 with
      A(f,fargs),A(g,gargs) ->
          if f <> g then raise Unify else
          itlist2 (fol_unify offset) fargs gargs sofar
    | _,V(x) ->
         (let x' = x + offset in
          try let tm2' = rev_assoc x' sofar in
              fol_unify 0 tm1 tm2' sofar
          with Not_found ->
              if istriv sofar x' tm1 then sofar
              else (tm1,x')::sofar)
    | V(x),_ ->
         (try let tm1' = rev_assoc x sofar in
              fol_unify offset tm1' tm2 sofar
          with Not_found ->
              let tm2' = fol_subst_bump offset [] tm2 in
              if istriv sofar x tm2' then sofar
              else (tm2',x)::sofar);;

let rec fol_subst_partial insts tm =
    match tm with
      V(v) -> (try let t = rev_assoc v insts in
                      fol_subst_partial insts t
                  with Not_found -> tm)
    | A(f,args) -> A(f,List.map (fol_subst_partial insts) args);;

let inst_lit inst (p, t) = (p, List.map (fol_subst_partial inst) t);;


let separate_insts offset oldinsts newinsts =
    let locins,globins =
      qpartition (fun (_,v) -> offset <= v) oldinsts newinsts in
    if globins = oldinsts then
      List.map (fun (t,x) -> fol_subst_partial newinsts t,x) locins,oldinsts
    else
      List.map (fun (t,x) -> fol_subst_partial newinsts t,x) locins,
      List.map (fun (t,x) -> fol_subst_partial newinsts t,x) globins;;

(*let unify_list env l1 l2 = List.fold_left2 unify env l1 l2;;*)

(*let unify_lit env ((h1 : int), l1) (h2, l2) =
  if h1 <> h2 then raise Unify else List.fold_left2 unify env l1 l2;;*)

let rec forall2 p l1 l2 =
  match (l1,l2) with
    [],[] -> true
  | (h1::t1,h2::t2) -> p h1 h2 & forall2 p t1 t2
  | _ -> false;;

let rec eq insts tm1 tm2 =
    tm1 == tm2 or
    match tm1,tm2 with
      A(f,fargs),A(g,gargs) ->
          f = g & forall2 (eq insts) fargs gargs
    | _,V(x) ->
         (try let tm2' = rev_assoc x insts in
              eq insts tm1 tm2'
          with Not_found ->
          try istriv insts x tm1 with Unify -> false)
    | V(x),_ ->
         (try let tm1' = rev_assoc x insts in
              eq insts tm1' tm2
          with Not_found ->
          try istriv insts x tm2 with Unify -> false);;

let eq_lit insts (p1,args1) (p2,args2) =
    p1 = p2 & forall2 (eq insts) args1 args2;;

let rec rev_itlist2 f l1 l2 b =
   match (l1,l2) with
     ([],[]) -> b
   | (h1::t1,h2::t2) -> rev_itlist2 f t1 t2 (f h1 h2 b)
      | _ -> failwith "rev_itlist2";;

let unify_rename_subst off l1 l2 sub list =
  let allins = rev_itlist2 (fol_unify off) l1 l2 sub in
  let locin,globin = separate_insts off sub allins in
  let mk_ihyp h = fol_inst_bump off locin h in
  let nrest =  List.map mk_ihyp list in
  globin, nrest;;

let unify_lit sub (p1, l1) (p2, l2) =
  if p1 <> p2 then raise Unify else
  rev_itlist2 (fol_unify 0) l1 l2 sub
;;

let unify_list sub l1 l2 = rev_itlist2 (fol_unify 0) l1 l2 sub;;

let empty_sub = ([] : (term * int) list);;

module Im = Map.Make(struct type t = int let compare = compare end);;
