type term = V of int
          | A of int * int * term list;;

exception Unify;;

let subst = Array.create 200000 (None : term option);;
let sh = ref [];;

let rec istriv x o = function
  | V y -> y + o = x || (match subst.(y + o) with Some t -> istriv x 0 t | None -> false)
  | A (f, oa, a) -> List.exists (istriv x (o + oa)) a && raise Unify;;

let rec bump_small off = function
  | V v -> V(v + off)
  | A(f, o, a) -> A(f, o + off, a);;

let rec unify o1 o2 tm1 tm2 =
  match tm1, tm2 with
  | A(f, o1a, fargs), A(g, o2a, gargs) ->
     if f <> g then raise Unify
     else List.iter2 (unify (o1 + o1a) (o2 + o2a)) fargs gargs
  | tm, V(x) -> (let x = x + o2 in match subst.(x) with
    | Some t -> unify o1 0 tm t
    | None -> if istriv x o1 tm then () else (subst.(x) <- Some (bump_small o1 tm); sh := x :: !sh))
  | V(x), tm -> (let x = x + o1 in match subst.(x) with
    | Some t -> unify 0 o2 t tm
    | None -> if istriv x o2 tm then () else (subst.(x) <- Some (bump_small o2 tm); sh := x :: !sh));;

let rec unify_rename off t1 t2 = unify 0 off t1 t2;;

let unify tm1 tm2 = unify 0 0 tm1 tm2;;

let rec restore_subst env =
  if not (!sh == env) then
    (subst.(List.hd !sh) <- None; sh := List.tl !sh; restore_subst env);;

let unify_list env l1 l2 = restore_subst env; List.iter2 unify l1 l2; !sh;;

let unify_lit env ((h1 : int), l1) (h2, l2) =
  if h1 <> h2 then raise Unify else unify_list env l1 l2;;

let rec eq2 x o = function
  | V y -> y + o = x || (match subst.(y + o) with Some t -> eq2 x 0 t | None -> false)
  | A (_, _, _) -> false;;

let rec eq o1 o2 tm1 tm2 =
(*  tm1 == tm2 ||*)
  match tm1,tm2 with
    A(f,o1a,fargs),A(g,o2a,gargs) -> f = g && List.for_all2 (eq (o1+o1a) (o2+o2a)) fargs gargs
  | _,V(x) -> let x = x + o2 in (match subst.(x) with Some t -> eq o1 0 tm1 t | None -> eq2 x o1 tm1)
  | V(x),_ -> let x = x + o1 in (match subst.(x) with Some t -> eq 0 o2 t tm2 | None -> eq2 x o2 tm2);;
let eq = eq 0 0;;

let eq_lit env (p1,args1) (p2,args2) =
  p1 = p2 && (restore_subst env; List.for_all2 eq args1 args2);;

let unify_rename_subst off l1 l2 env list =
  restore_subst env;
  List.iter2 (unify_rename off) l1 l2;
  !sh, List.map (fun (p,l) -> p,List.map (bump_small off) l) list;;

let rec inst_tm o = function
    V(v) -> (match subst.(v+o) with Some t -> inst_tm 0 t | None -> V(v+o))
  | A(f,oa,args) -> A(f,0,List.map (inst_tm (o+oa)) args);;
let inst_lit env (p, l) = restore_subst env; (p, List.map (inst_tm 0) l);;

(* In leanCoP: only for printing, in resolve is used by unify_rename2 *)
(*

let unify_rename2 off l1 l2 rest1 rest2 =
  restore_subst [];
  List.iter2 (unify_rename off) l1 l2;
  ((List.map (fun (p,l) -> p, List.map inst_tm l) rest1),
  (List.map (fun (p,l) -> p, List.map (fun t -> inst_tm (bump_subst_vars off t)) l) rest2));;
 *)

let empty_sub = [];;

module Im = Map.Make(struct type t = int let compare = compare end);;
