type term = V of int
          | A of int * term list;;

exception Unify;;

(* Substitutions are maps from int to term *)
module Im = Map.Make(struct type t = int let compare = compare end);;

let rec istriv env x = function
    V y -> y = x || (try istriv env x (Im.find y env) with Not_found -> false)
  | A (f, a) -> List.exists (istriv env x) a && raise Unify;;

let rec unify env tm1 tm2 = match tm1,tm2 with
    A(f,fargs),A(g,gargs) ->
      if f <> g then raise Unify else
        List.fold_left2 unify env fargs gargs
  | _,V(x) ->
     (try unify env tm1 (Im.find x env)
     with Not_found -> if istriv env x tm1 then env else Im.add x tm1 env)
  | V(x),_ ->
     try unify env (Im.find x env) tm2
     with Not_found -> if istriv env x tm2 then env else Im.add x tm2 env;;

let unify_list env l1 l2 = List.fold_left2 unify env l1 l2;;

let unify_lit env ((h1 : int), l1) (h2, l2) =
  if h1 <> h2 then raise Unify else List.fold_left2 unify env l1 l2;;

(*let offinc = 10000;;*)

let rec bump_small off tm = match tm with
    V v -> V(v + off)
  | A(f, a) -> A(f, List.map (bump_small off) a);;

(* Unification with renaming of the second argument *)
let rec unify_rename goff off ((env, (nv1, nv2)) as sf) tm1 tm2 = match tm1,tm2 with
    A(f,fargs),A(g,gargs) ->
      if f <> g then raise Unify else
        List.fold_left2 (unify_rename goff off) sf fargs gargs
  | _,V(x) -> let x = x + off in
     (try unify_rename goff 0 sf tm1 (Im.find x env)
     with Not_found -> if istriv env x tm1 then sf else (Im.add x tm1 env, if x < goff then ((x,tm1) :: nv1, nv2) else (nv1, (x,tm1) :: nv2)))
  | V(x),_ ->
     try unify_rename goff off sf (Im.find x env) tm2
     with Not_found ->
       let tm2' = bump_small off tm2 in
       if istriv env x tm2' then sf else (Im.add x tm2' env, if x < goff then ((x,tm2')::nv1, nv2) else (nv1, (x,tm2')::nv2));;

let rec eq2 env x = function
    V y -> y = x || (try eq2 env x (Im.find y env) with Not_found -> false)
  | A (f, a) -> false;;

let rec eq env tm1 tm2 =
  tm1 == tm2 ||
  match tm1,tm2 with
    A(f,fargs),A(g,gargs) -> f = g && List.for_all2 (eq env) fargs gargs
  | _,V(x) ->
    (try eq env tm1 (Im.find x env)
     with Not_found -> eq2 env x tm1)
  | V(x),_ ->
    (try eq env (Im.find x env) tm2
     with Not_found -> eq2 env x tm2);;

let eq_lit env (p1,args1) (p2,args2) =
  p1 = p2 && List.for_all2 (eq env) args1 args2;;

(* Only for printing *)
let rec inst_tm env tm = match tm with
    V(v) -> (try inst_tm env (Im.find v env) with Not_found -> tm)
  | A(f,args) -> A(f,List.map (inst_tm env) args);;
let inst_lit env (p, l) = (p, List.map (inst_tm env) l);;

let rec bump_vars off = function
    V x -> V (x + off)
  | A (x, l) -> A (x, List.map (bump_vars off) l);;

let rec subst_vars list = function
    V x -> (try subst_vars list (List.assoc x list) with Not_found -> V x)
  | A (x, l) -> A (x, List.map (subst_vars list) l);;

let rec bump_subst_vars off list = function
    V x -> subst_vars list (V (x + off))
  | A (x, l) -> A (x, List.map (bump_subst_vars off list) l);;

let unify_rename_subst off l1 l2 sub list =
  let (s, (nv1, nv2)) = List.fold_left2 (unify_rename off off) (sub, ([], [])) l1 l2 in
    (*Printf.printf "%i \t\t[%s]\t[%s]\n" off (String.concat "," (List.map string_of_int (List.map fst nv1))) (String.concat "," (List.map string_of_int (List.map fst nv2)));*)
  match nv1, nv2 with
    [], [] -> (sub, List.map (fun (p,l) -> (p,List.map (bump_vars off) l)) list)
  | [], _  -> (sub, List.map (fun (p,l) -> (p,List.map (bump_subst_vars off nv2) l)) list)
  | _ , [] -> (s,   List.map (fun (p,l) -> (p,List.map (bump_vars off) l)) list)
  | _ , _  ->
     (List.fold_left (fun sf (v,t) -> Im.add v (subst_vars nv2 t) sf) sub nv1,
     List.map (fun (p,l) -> (p,List.map (bump_subst_vars off nv2) l)) list)

let empty_sub = Im.empty;;
