open Cnf;;

module I2m = Map.Make(struct type t = (int * int) let compare = compare end);;
let i2m_find k m = try Some (I2m.find k m) with Not_found -> None;;
type 'a dt = Node of ('a dt * 'a dt list) I2m.t * 'a list;;

let empty_dt = Node (I2m.empty, []);;

(*let rec insert nv (Node(map, v, _)) = function
| [] -> Node(map, nv :: v, [])
| (A (f, args) :: rest) ->
    let l = List.length args in
    let child = try I2m.find (f, l) map with Not_found -> empty_dt in
    let new_child = insert nv child (args @ rest) in
    Node (I2m.add (f, l) new_child map, v, [])
| (V _ :: rest) ->
    let child = try I2m.find (0, 0) map with Not_found -> empty_dt in
    let new_child = insert nv child rest in
    Node (I2m.add (0, 0) new_child map, v, []);;

let rec find_jl skip (_, n) (Node (map, _, _) as ch) acc =
  let skip = skip + n - 1 in
  if skip = 0 then (if List.mem ch acc then acc else ch :: acc) else
  I2m.fold (find_jl skip) map acc;;

let rec update_jl (Node (map, v, _)) =
  let map = I2m.map update_jl map in
  Node (map, v, List.sort compare (I2m.fold (find_jl 1) map []));;*)

let rec drop n = function
    [] -> []
  | ((_ :: t) as l) -> if n = 0 then l else drop (n-1) t;;

let rec insert nv (Node(map, v)) = function
| [] -> let l = Node(map, nv :: v) in l, [l]
| (h :: rest) ->
    let f, args, l = match h with
      A (f, args) -> f, args, List.length args
    | V _ -> 0, [], 0 in
    let child, jumps = try I2m.find (f, l) map with Not_found -> empty_dt, [] in
    let nchild, jstack = insert nv child (args @ rest) in
    let nstack = drop l jstack in
    let njump = List.hd nstack in
    let jumps = if List.mem njump jumps then jumps else njump :: jumps in
    let node = Node (I2m.add (f, l) (nchild, jumps) map, v) in
    node, node :: nstack;;
let insert nv dt ts = fst (insert nv dt ts);;

let rec mem nv (Node(map, v)) = function
| [] -> List.mem nv v
| (A (f, args) :: rest) ->
    let l = List.length args in
    (try mem nv (fst (I2m.find (f, l) map)) (args @ rest) with Not_found -> false)
| (V _ :: rest) ->
    (try mem nv (fst (I2m.find (0, 0) map)) rest with Not_found -> false);;

let generalize tm net =
  let rec look acc (Node(map, v)) = function
      [] -> List.fold_left (fun sf x -> x :: sf) acc v
    | (A (f, args) :: rest) ->
        let l = List.length args in
        let acc =
          begin match i2m_find (0, 0) map with
          | None -> acc
          | Some (ch, _) -> look acc ch rest
          end in
        begin match i2m_find (f, l) map with
        | None -> acc
        | Some (ch, _) -> look acc ch (args @ rest)
        end
    | (V _ :: rest) ->
        match i2m_find (0, 0) map with
        | None -> acc
        | Some (ch, _) -> look acc ch rest in
  look [] net [tm];;

let rec unifs acc (Node(map, v)) = function
  [] -> List.fold_left (fun sf x -> x :: sf) acc v
| (A (f, args) :: rest) ->
    let l = List.length args in
    let acc =
      begin match i2m_find (0, 0) map with
      | None -> acc
      | Some (ch, _) -> unifs acc ch rest
      end in
    begin match i2m_find (f, l) map with
    | None -> acc
    | Some (ch, _) -> unifs acc ch (args @ rest)
    end
| (V _ :: rest) ->
    I2m.fold (fun _ (_, jl) acc -> List.fold_left (fun sf n -> unifs sf n rest) acc jl) map acc;;

let rec trace_unifs acc ((Node(map, v)) as dt) = function
  [] -> List.fold_left (fun sf x -> x :: sf) acc v
| (A (f, args) :: rest) ->
    let l = List.length args in
    let acc =
      begin match i2m_find (0, 0) map with
      | None -> acc
      | Some (ch, _) -> trace_unifs acc ch rest
      end in
    begin match i2m_find (f, l) map with
    | None -> acc
    | Some (ch, _) -> trace_unifs acc ch (args @ rest)
    end
| (V i :: rest) ->
    match subst.(i) with
      Some t -> trace_unifs acc dt (t :: rest)
    | None -> I2m.fold (fun _ (_, jl) acc -> List.fold_left (fun sf n -> unifs sf n rest) acc jl) map acc;;

let trace_unifs sub dt tms = restore_subst sub; trace_unifs [] dt tms;;

(*
open Cnf;;
open Dtree;;

#install_printer Fof.pp_print_term;;
Hashtbl.add Fof_parse.no_cnst 1 "a";;
Hashtbl.add Fof_parse.no_cnst 2 "b";;
Hashtbl.add Fof_parse.no_cnst 3 "c";;
Hashtbl.add Fof_parse.no_cnst 6 "f";;
Hashtbl.add Fof_parse.no_cnst 7 "g";;

let a =  A(1, []);;
let b =  A(2, []);;
let c =  A(3, []);;
let gab = A(7, [a; b]);;
let gax = A(7, [a; V 1]);;
let gxc = A(7, [V 1; c]);;
let gxb = A(7, [V 1; b]);;
let gbc = A(7, [b; c]);;

let t1 = A(6, [gax; c]);;
let t2 = A(6, [gxb; V 2]);;
let t3 = A(6, [gab; a]);;
let t4 = A(6, [gxc; b]);;
let t5 = A(6, [V 3; V 4]);;
let t6 = A(6, [gbc; V 2]);;

let dt = (List.fold_left (fun sf x -> insert x sf [x]) empty_dt [t1; t2; t3; t4; t5; t6]);;

let gbx = A(7, [b; V 5]);;
let t = A(6, [gbx; V 0]);;

generalize t dt;;
unifs [] dt [t];;
*)
