open Cnf;;
open Fof_parse;;
open Fof;;
open Features;;

let hash = pred_number "'#'";;
let hasht = Atom(hash, []);;
let db = Hashtbl.create 10017;;

(* DB entries are pairs of the flat info:
    (term list * (int * term list) list * int * int) list
  and possibly the learned info indexed by problem features:
    (int, ((term list * (int * term list) list * int * int) list) Hashtbl.t)
  with size 0 if nothing is known *)

let conj = true;;
let cut1 = true;;
let cut2 = true;;
let cut3 = ref true;;
let cut4 = ref false;;
let filterdisjont = false;;

let infer = ref 0;;
let depth = ref 0;;
let lem = ref 0;;
let pat = ref 0;;
let sub = ref 0;;

exception Cut of int;;

let unify_rename_by (sub, off) args1 (args2, rest, vars) =
  if vars = 0 then ((unify_list sub args1 args2, off), rest) else
  let s, rest = unify_rename_subst off args1 args2 sub rest in ((s, off + vars), rest);;

let catch_cut n f arg = try f arg with (Cut m) -> if n = m then () else raise (Cut m);;

type proof = Lem of lit | Pat of lit | Res of (lit * lit list * lit list * int);;
exception Solved of int list * proof list

let learned = (*try
    let ic = open_in "./datai" in
    let learned = ((input_value ic) : (int, int list) Hashtbl.t) in
    close_in ic; learned
  with _ -> *)Hashtbl.create 1;;

let empty_learned = Array.make 0 (([], [], 0, 0), 0.0), Hashtbl.create 1;;

let knn db (arr, hash) feas = match db with
  [] -> []
| [h] -> db
| _ when Hashtbl.length hash = 0 -> db
| _ ->
  let arr = Array.copy arr in
  Im.iter (fun f () -> List.iter (fun (i, wei) -> let (v,sf) = arr.(i) in arr.(i) <- (v, sf +. wei)) (Hashtbl.find_all hash f)) feas;
(*  Array.iter (fun (_, v) -> Format.printf "%.3f " v) arr;
  Format.printf "\n";*)
  Array.stable_sort (fun (_, v1) (_, v2) -> compare v2 v1) arr;
  List.map fst (Array.to_list arr)
;;

(*let print_clause subst clause = print_clause (List.map (inst_lit subst) clause);;*)

let rec prove (sub, off) (path, fea) lim lem stack = function
  [] -> begin match stack with
      [] -> raise (Solved ([],[])) (* TODO: return sub? *)
    | (cut, p, lim, l, c) :: t ->
        if !cut3 then (prove (sub, off) p lim l t c; raise (Cut cut)) else prove (sub, off) p lim l t c
    end
| ((lit1 :: cla) as clause) ->
    (*print_clause sub clause; Format.print_string "\t\t"; print_clause sub (List.rev path); Format.print_char '\n';*)

    if not ((List.exists (fun x -> List.exists (eq_lit sub x) path)) clause) then begin

    if List.exists (eq_lit sub lit1) lem && ((try prove (sub, off) (path, fea) lim lem stack cla with Solved (s,x) -> raise (Solved (s,Lem lit1 :: x))); cut1) then () else

    let neglit = negate lit1 in

    if List.fold_left (fun sf plit -> if sf then true else
      try let sub = unify_lit sub neglit plit in
      ((try prove (sub, off) (path, fea) lim (lit1 :: lem) stack cla with Solved (s,x) -> raise (Solved (s,Pat lit1 :: x))); cut2) with Unify -> sf) false path then () else

    let db, learned = try Hashtbl.find db (fst neglit) with Not_found -> ([], empty_learned) in
    let nfea = fea in
  (*let nfea = if db = [] then Im.empty else featuresl fea lit1 in
    let db = knn db learned nfea in*)

    let fn (s, cla1) hsh =
      infer := !infer + 1;
      let stack = ((if !cut3 then lim else min_int), (path, fea), lim, lit1 :: lem, cla) :: stack in
      try prove s (lit1 :: path, nfea) (lim - 1) lem stack cla1; if !cut4 then raise (Cut lim) with Solved (s,x) -> raise (Solved (s,Res (lit1, path, lem, hsh) :: x)) in
    let iter_fun (lit2, cla2, vars, hsh) = if lim > 0 || vars = 0 then try fn (unify_rename_by (sub, off) (snd lit1) (lit2, cla2, vars)) hsh with Unify -> () in
    if !cut3 || !cut4 then catch_cut lim (List.iter iter_fun) db else List.iter iter_fun db end;;

let prove lim = prove (empty_sub, 0) ([], Im.empty) lim [] [] [(hash,[])];;

let indexed = ref 0;;
let indexedknown = ref 0;;

let axioms2db all_fea axs =
  let rec iter_rest acc fnctn = function
    [] -> ()
  | h :: t -> fnctn h (List.filter (fun x -> x <> (-hash, []))
                         (List.rev_append acc t)); iter_rest (h :: acc) fnctn t in
  let predb = Hashtbl.create 100 in
  let cl2predb cl =
    let max_var = 1 + List.fold_left (fun sf (_, t) -> List.fold_left max_var sf t) (-1) cl in
    let cl = if not conj && List.for_all (fun (p, _) -> p < 0) cl &&
      not (List.mem (-hash, []) cl) then (-hash, []) :: cl else cl in
    let contr_hash lit rest = contr_number (string_of_contr lit rest) in
    iter_rest [] (fun (p,tl) rest -> Hashtbl.add predb p
      (tl, rest, max_var, contr_hash (p, tl) rest)) cl in
  List.iter cl2predb axs;
  Hashtbl.clear db;
  let cl2db k _ =
    let v = List.rev (Hashtbl.find_all predb k) in
    let glcommon (_,_,_,hsh) =
      List.fold_left (List.fold_left (fun sf x -> if Im.mem x all_fea then Im.add x () sf else sf))
        Im.empty (Hashtbl.find_all learned hsh) in
    incr indexed;
    Hashtbl.add db k (
      v, if List.for_all (fun x -> glcommon x = Im.empty) v then empty_learned else begin
        incr indexedknown;
        let v2 = if filterdisjont then List.filter (fun e -> glcommon e <> Im.empty) v else v in

        let reta = Array.of_list (List.map (fun x -> (x, 0.0)) v2) in
        let reth = Hashtbl.create 100 in
        let positions = Hashtbl.create 100 in
        Array.iteri (fun i v -> Hashtbl.add positions (fst v) i) reta;

        (* The NOTFIDF version *)
        let addfea f e = Hashtbl.add reth f (Hashtbl.find positions e, 1.0) in
        List.iter (fun e -> Im.iter (fun k () -> addfea k e) (glcommon e)) v2;

        (* TFIDF *)
(*        let thn = log (float_of_int (List.fold_left (+) 1 (List.map (fun (_,_,_,hsh) -> List.length (Hashtbl.find_all learned hsh)) v2))) in
        let ffreq = Hashtbl.create 100 in
        let addffreq i =
          try Hashtbl.replace ffreq i (Hashtbl.find ffreq i + 1)
          with Not_found -> Hashtbl.add ffreq i 1 in
        List.iter (fun (_,_,_,hsh) -> List.iter (List.iter addffreq) (Hashtbl.find_all learned hsh)) v2;
        let addfea f e = Hashtbl.add reth f (Hashtbl.find positions e, thn -. log (float_of_int (Hashtbl.find ffreq f))) in
        List.iter (fun ((_,_,_,c) as e) -> List.iter (List.iter (fun (f : int) -> addfea f e)) (Hashtbl.find_all learned c)) v2;*)

        reta, reth
      end
    )
  in
  let keys = Hashtbl.fold (fun k _ sf -> Im.add k () sf) predb Im.empty in
  Im.iter cl2db keys
;;

let file_mat file =
  let (ths, gl) = Fof_lexer.problem file in
  let forms = equal_axioms (Neg(Conj(hasht, gl)) :: ths) in
  let cnf = List.map
    (fun f -> cnf (noforall (skolem (nnf (rename_form (miniscope f)))))) forms in
  List.fold_left matrix [] cnf
;;

let leancop file =
  let mat = file_mat file in
  let all_fea = List.fold_left (List.fold_left featuresl) Im.empty ([(hash, [])] :: mat) in
(*  List.iter (fun x -> Format.print_char '%'; print_clause x; Format.print_char '\n') mat; Format.printf "%%--\n%!";*)
  axioms2db all_fea mat;
  (*maxsec 10*) (fun () -> for i = 0 to 1000 do depth := i; prove i done) ();
  failwith "over 1000";
;;

exception Alarm;;
exception CleanExit;;
let alarm _ = raise Alarm;;
let cleanexit _ = raise CleanExit;;
Sys.signal Sys.sigint (Sys.Signal_handle cleanexit);;
Sys.signal Sys.sigterm (Sys.Signal_handle cleanexit);;
Sys.signal Sys.sigalrm (Sys.Signal_handle alarm);;
let maxsec i f a = ignore (Unix.alarm i); try f a; ignore (Unix.alarm 0); () with Alarm -> ();;
at_exit (fun () ->
Format.printf "%% Inf: %i Depth: %i Indexed: %i Known: %i\n" !infer !depth !indexed !indexedknown);;

let print_solved_features pl =
  let rec f2f acc = function
    Res (lit, path, lem, contr) :: t -> f2f ((lit, path, lem, contr) :: acc) t
  | _ :: t -> f2f acc t
  | [] -> acc
  in
  let l = f2f [] pl in
  List.iter (fun (lit, path, lem, contr) -> Format.print_string (Hashtbl.find no_contr contr); Format.print_string ", "; pp_print_lit Format.std_formatter lit; Format.print_string ", ("; pp_iter Format.std_formatter pp_print_lit "," path; Format.print_string "), ("; pp_iter Format.std_formatter pp_print_lit "," lem; Format.print_string ").\n") l;;



if Array.length Sys.argv > 1 then
  try leancop (Sys.argv.(1))
  with Solved (_, pl) -> Format.printf "\n%% SZS status Theorem\n%!"; print_solved_features pl
    | Failure x -> Format.printf "\n%% SZS status Error\n%%%s\n%!" x
    | CleanExit -> Format.printf "\n%% SZS status Unknown\n%!"
    | Parsing.Parse_error -> Format.printf "\n%% SZS status Error\n%%Parse_error\n%!";;
