open Cnf;;
open Fof_parse;;
open Fof;;
open Features;;

let hash = pred_number "'#'";;
let hasht = Atom(hash, []);;
(*let db = ref ([] : (int * (term list * (int * term list) list * int) list) list);;*)
let db = Hashtbl.create 10017;;

let conj = true;;
let cut1 = true;;
let cut2 = true;;
let cut3 = true;;
let def = false;;

let infer = ref 0;;
let depth = ref 0;;

let eq (sub, off) l1 l2 = eq_lit sub l1 l2;;

let unify (sub, off) l1 l2 = try Some (unify_lit sub l1 l2, off) with Unify -> None;;

let unify_rename (s, off) args1 (args2, rest, vars, _) =
  try Some(if vars = 0 then ((unify_list s args1 args2, off), rest) else
    let s, rest = unify_rename_subst off args1 args2 s rest in ((s, off + vars), rest))
  with Unify -> None;;

(* print_clause (fst sub) (lit1 :: rest); print_string "\t\t"; print_clause (fst sub) (List.rev path); print_char '\n'; *)

type proof = Lem of lit | Pat of lit | Res of (lit * lit list * lit list * int);;
exception Solved of (int list * proof) list

(*let print_clause (subst, _) clause = print_clause (List.map (inst_lit subst) clause);;*)

let rec split_list = function
  h1 :: h2 :: t -> let (l1, l2) = split_list t in (h1 :: l1, h2 :: l2)
| l -> ([], l);;

let rec prove_one sub lit1 rest ((path, lem, lim) as hist) alt (todo, prf) =
  (*print_clause sub (lit1 :: rest); Format.print_string "\t\t"; print_clause sub (List.rev path); Format.print_char '\n';*)
  if (List.exists (fun x -> List.exists (eq sub x) path)) (lit1 :: rest) then alt () else
  if List.exists (eq sub lit1) lem then prove_many sub hist (if cut1 then alt else (fun () -> prove_path sub lit1 rest hist alt (todo, prf) (negate lit1) path)) (todo, (fst sub, Lem lit1) :: prf) rest else
  prove_path sub lit1 rest hist alt (todo, prf) (negate lit1) path

and prove_path sub lit1 rest hist alt (todo, prf) neglit = function
  | plit :: pt -> (match unify sub neglit plit with
    | Some sub2 -> prove_many sub2 hist (if cut2 then alt else (fun () -> prove_path sub lit1 rest hist alt (todo, prf) neglit pt)) (todo, (fst sub, Pat lit1) :: prf) rest
    | None -> prove_path sub lit1 rest hist alt (todo, prf) neglit pt)
  | [] -> prove_lit sub lit1 rest hist alt (todo, prf) (try Hashtbl.find db (fst neglit) with Not_found -> [])

and prove_lit sub lit1 rest ((path, lem, lim) as hist) alt (todo, prf) = function
  | ((_,nrest,vars,hsh) as eh) :: et -> (match if vars > 0 && lim - List.length nrest <= 0 then None else unify_rename sub (snd lit1) eh with
    | Some (sub2, cla1) ->
      let ntodo (sub, nalt, prf, nlim) = prove_many sub (path, lit1 :: lem, nlim) (if cut3 then alt else nalt) (todo, prf) rest in
      infer := !infer + 1; prove_many sub2 (lit1 :: path, lem, lim - 1) (fun () -> prove_lit sub lit1 rest hist alt (todo, prf) et) (ntodo, (fst sub, Res (lit1, path, lem, hsh)) :: prf) cla1
    | None -> prove_lit sub lit1 rest hist alt (todo, prf) et)
  | [] -> alt ()

(*and prove_many sub ((path, lem, lim) as hist) alt (todo, prf) = function
  | h :: t -> prove_one sub h t hist alt (todo, prf)
  | [] -> todo (sub, alt, prf, lim)*)

and prove_many sub ((path, lem, lim) as hist) alt (todo, prf) = function
  | [h] -> prove_one sub h [] hist alt (todo, prf)
  | ((_ :: _) as l) ->
      let (l1, l2) = split_list l in
      let todol2 (sub, nalt, prf, nlim) = prove_many sub (path, lem, nlim) nalt (todo, prf) l1 in
      let alt1 () = prove_many sub (path, lem, lim - lim / 3) alt (todol2, prf) l2 in
      let todol1 (sub, nalt, prf, nlim) = prove_many sub (path, lem, nlim) nalt (todo, prf) l2 in
      prove_many sub (path, lem, lim / 3) alt1 (todol1, prf) l1
  | [] -> todo (sub, alt, prf, lim)

let prove lim = prove_one (empty_sub, 0) (hash,[]) [] ([], [], lim) (fun () -> ()) ((fun (_,_,prf,_) -> raise (Solved prf)), []);;

let axioms2db axs =
  let rec iter_rest acc fnctn = function
    [] -> ()
  | h :: t -> fnctn h (List.filter (fun x -> x <> (-hash, []))
                         (List.rev_append acc t)); iter_rest (h :: acc) fnctn t in
  let predb = Hashtbl.create 100 in
  let cl2predb cl =
    let max_var = 1 + List.fold_left (fun sf (_, t) -> List.fold_left max_var sf t) (-1) cl in
    let cl = if not conj && List.for_all (fun (p, _) -> p < 0) cl &&
      not (List.mem (-hash, []) cl) then (-hash, []) :: cl else cl in
    let contr_hash lit rest = contr_number (string_of_contr lit rest) in
    iter_rest [] (fun (p,tl) rest -> Hashtbl.add predb p
      (tl, rest, max_var, contr_hash (p, tl) rest)) cl in
  List.iter cl2predb axs;
  Hashtbl.clear db;
  let cl2db k _ =
    let v = List.rev (Hashtbl.find_all predb k) in
    Hashtbl.add db k v in
  let keys = Hashtbl.fold (fun k _ sf -> Im.add k () sf) predb Im.empty in
  Im.iter cl2db keys
;;

let file_mat file =
  let (ths, gl) = Fof_lexer.problem file in
  let forms = equal_axioms (Neg(Conj(hasht, gl)) :: ths) in
  let precnf = List.map (fun f -> noforall (skolem (nnf (rename_form (miniscope f))))) forms in
  if def then List.fold_left dcnf [] precnf
  else List.fold_left matrix [] (List.map cnf precnf)
;;

let leancop file =
  let mat = file_mat file in
  (*List.iter (fun x -> Format.print_char '%'; print_clause x; Format.print_char '\n') mat; Format.printf "%%--\n%!";*)
  axioms2db mat;
  for i = 0 to 1000 do depth := i; prove i done;
  failwith "over 1000"
;;

exception Alarm;;
exception CleanExit;;
let alarm _ = raise Alarm;;
let cleanexit _ = raise CleanExit;;
Sys.signal Sys.sigint (Sys.Signal_handle cleanexit);;
Sys.signal Sys.sigterm (Sys.Signal_handle cleanexit);;
Sys.signal Sys.sigalrm (Sys.Signal_handle alarm);;
let maxsec i f a = ignore (Unix.alarm i); try f a; ignore (Unix.alarm 0); () with Alarm -> ();;
at_exit (fun () -> Format.printf "%%Inf: %i Depth: %i\n" !infer !depth);;

let print_solved_features pl =
  let rec f2f acc = function
    (s, Res (lit, path, lem, contr)) :: t -> f2f ((inst_lit s lit, List.map (inst_lit s) path, List.map (inst_lit s) lem, contr) :: acc) t
  | _ :: t -> f2f acc t
  | [] -> acc
  in
  let l = f2f [] pl in
  List.iter (fun (lit, path, lem, contr) -> Format.print_string (Hashtbl.find no_contr contr); Format.print_string ", "; pp_print_lit Format.std_formatter lit; Format.print_string ", ("; pp_iter Format.std_formatter pp_print_lit "," path; Format.print_string "), ("; pp_iter Format.std_formatter pp_print_lit "," lem; Format.print_string ").\n") l;;

if Array.length Sys.argv > 1 then
  try leancop (Sys.argv.(1))
  with Solved prf -> Format.printf "\nSZS status Theorem\n%!"; print_solved_features prf
     | Failure x -> Format.printf "\n%% SZS status Error\n%%%s\n%!" x
     | CleanExit -> Format.printf "\n%% SZS status Unknown\n%!"
     | Parsing.Parse_error -> Format.printf "\n%% SZS status Error\n%%Parse_error\n%!";;
