open Fof_parse;;
open Cnf;;
open Fof;;
open Arglean;;

let db = Hashtbl.create 10017;;

let eq (sub, off) l1 l2 = eq_lit sub l1 l2;;

let unify (sub, off) l1 l2 = try Some (unify_lit sub l1 l2, off) with Unify -> None;;

let unify_rename (s, off) args1 (args2, rest, vars, _) =
  try Some(if vars = 0 then ((unify_list s args1 args2, off), rest) else
    let s, rest = unify_rename_subst off args1 args2 s rest in ((s, off + vars), rest))
  with Unify -> None;;

exception Solved;;

let print_clause (subst, _) clause = print_clause (List.map (inst_lit subst) clause);;

open List;;
let rec drop n l = if n = 0 then l else (drop (n-1) (tl l));;

let indexed = ref 0;;
let infer = ref 0;;
let depth = ref 0;;
let msubl = ref 0;;
let mpatl = ref 0;;
let mleml = ref 0;;
let maltl = ref 0;;
let mtodl = ref 0;;
let moffs = ref 0;;
let updmax r l = let l = length l in if l > !r then r := l;;

let prove pathlim =
  let sub = ref ([],0) and cl = ref ([hash, []], 0) and path = ref []
  and lem = ref [] and alt = ref [] and todo = ref [] in
  let rec fone () =
    updmax msubl (fst !sub);
    updmax mpatl !path;
    updmax mleml !lem;
    updmax maltl !alt;
    updmax mtodl !todo;
    if snd !sub > !moffs then moffs := snd !sub;
    if length (fst !cl) <= snd !cl then do_todo () else
    (let thiscl = drop (snd !cl) (fst !cl) in
     (*print_clause !sub thiscl; Format.print_string "\t\t"; print_clause !sub (rev !path);
     Format.print_string "\t\t"; Format.print_int (length !todo); Format.print_char '\n';*)
     if (exists (fun x -> exists (eq !sub x) !path)) thiscl then do_alt () else
    let lit1 = nth (fst !cl) (snd !cl) in
    if exists (eq !sub lit1) !lem then (cl := (fst !cl, snd !cl + 1); fone ()) else
    fpath 0)
  and fpath n =
    if n >= length !path then flit 0 else
    let plit = nth !path n in
    let neglit = negate (nth (fst !cl) (snd !cl)) in
    match unify !sub neglit plit with
    | Some sub2 ->
       if not !cut2(*or g*) then alt := (false, n + 1, !cl, length !path, length !lem, length !todo, !sub) :: !alt;
       cl := (fst !cl, snd !cl + 1); sub := sub2; fone ()
    | None -> fpath (n + 1)
  and flit n =
    let lit1 = nth (fst !cl) (snd !cl) in let neglit = negate lit1 in
    let dbs = try Hashtbl.find db (fst neglit) with Not_found -> [] in
    if n >= length dbs then do_alt () else
    let (((_,_,vars,hsh) as eh), chwei) = nth dbs n in
    if length !path >= pathlim && vars > 0 then flit (n + 1) else
    match unify_rename !sub (snd lit1) eh with
    | Some (sub2, cla1) -> infer := !infer + 1;
       todo := ((fst !cl, snd !cl + 1), length !path, length !lem, lit1, length !alt) :: !todo;
       alt := (true, n + 1, !cl, length !path, length !lem, length !todo - 1, !sub) :: !alt;
       sub := sub2; cl := (cla1, 0); path := lit1 :: !path; fone()
    | None -> flit (n + 1)
  and do_todo () =
    match !todo with
    | [] -> true
    | ((cls, cln), plen, llen, newlem, altlen) :: tr ->
       if !cut3 then alt := drop (length !alt - altlen) !alt;
       path := drop (length !path - plen) !path;
       lem := newlem :: drop (length !lem - llen) !lem;
       todo := tr; cl := (cls, cln); fone ()
  and do_alt () =
    match !alt with
    | [] -> false
    | (is_lit, n, ncl, plen, llen, tlen, nsub) :: at ->
       alt := at; sub := nsub; cl := ncl;
       path := drop (length !path - plen) !path;
       lem := drop (length !lem - llen) !lem;
       todo := drop (length !todo - tlen) !todo;
       if is_lit then flit n else fpath n
  in
  fone ();;

let rec iter_rest acc fnctn = function
  [] -> ()
| h :: t -> fnctn h (List.filter (fun x -> x <> (-hash, []))
             (List.rev_append acc t)); iter_rest (h :: acc) fnctn t;;

let axioms2db  axs =
  let predb = Hashtbl.create 100 in
  let cl2predb cl =
    let max_var = 1 + List.fold_left (fun sf (_, t) -> List.fold_left max_var sf t) (-1) cl in
    let cl = if not !conj && List.for_all (fun (p, _) -> p < 0) cl &&
      not (List.mem (-hash, []) cl) then (-hash, []) :: cl else cl in
    iter_rest [] (fun (p,tl) rest -> Hashtbl.add predb p
      (tl, rest, max_var, 0)) cl in
  List.iter cl2predb axs;
  Hashtbl.clear db;
  let cl2db k _ =
    let v = List.rev (Hashtbl.find_all predb k) in
    incr indexed;
    let vd = List.map (fun ((_, _, _, c) as e) -> (e, 0)) v in
    Hashtbl.add db k vd;
  in
  let keys = Hashtbl.fold (fun k _ sf -> Im.add k () sf) predb Im.empty in
  Im.iter cl2db keys
;;

let leancop file =
  let mat = file_mat true !def true file in
(*  List.iter (fun x -> Format.print_char '%'; print_clause x; Format.print_char '\n') mat; Format.printf "%%--\n%!";*)
  axioms2db mat;
  for i = 0 to 1000 do depth := i; if prove i then raise Solved; done;
  failwith "over 1000"
;;

exception Alarm;;
exception CleanExit;;
let alarm _ = raise Alarm;;
let cleanexit _ = raise CleanExit;;
Sys.signal Sys.sigint (Sys.Signal_handle cleanexit);;
Sys.signal Sys.sigterm (Sys.Signal_handle cleanexit);;
Sys.signal Sys.sigalrm (Sys.Signal_handle alarm);;
let maxsec i f a = ignore (Unix.alarm i); try f a; ignore (Unix.alarm 0); () with Alarm -> ();;
at_exit (fun () ->
Format.printf "%% Inf: %i Depth: %i Indexed: %i MSubL: %i MPatL: %i MLemL:%i MAltL:%i MTodL:%i MOff: %i\n" !infer !depth !indexed !msubl !mpatl !mleml !maltl !mtodl !moffs);;



  
if !tosolve <> "" then try leancop !tosolve
  with Solved -> Format.printf "\n%% SZS status Theorem\n%!"
     | Failure x -> Format.printf "\n%% SZS status Error\n%%%s\n%!" x
     | CleanExit -> Format.printf "\n%% SZS status Unknown\n%!"
     | Parsing.Parse_error -> Format.printf "\n%% SZS status Error\n%%Parse_error\n%!";;
