open Fol;;

Fusion.new_constant ("#", Fusion.bool_ty);;
let hash = Fusion.mk_const ("#", []);;
let hashfn = Fol.fol_of_literal true [] [] (Bool.mk_neg hash);;
let hashf = Fol.fol_of_literal true [] [] hash;;

let db = ref ([] : (int * (fol_term list * fol_lit list * bool) list) list);;

let inc_vars subst off (x, l) =
  let rec inc_vs = function
    Fvar x -> fol_subst_partial subst (Fvar (x + off))
  | Fnapp (x, l) -> Fnapp (x, List.map inc_vs l) in
  (x, List.map inc_vs l);;

(* Immediately substitute new introduced and substituted variables *)
let optimize_subst off (nsub, nv) rest =
  if nv = [] then (nsub, List.map (inc_vars Im.empty off) rest) else
  let nrest = List.map (inc_vars nsub off) rest in
  let nnsub = Im.filter (fun k _ -> k < off or not (List.mem k nv)) nsub in
  let nnsub = Im.mapi (fun k t -> ((if List.mem k nv then fol_subst_partial nsub t else t))) nnsub in
  nnsub, nrest
;;

let fol_unify_l off l1 l2 sf = List.fold_left (fun sf (t1, t2) -> fol_unify off t1 t2 sf) sf (Lib.zip l1 l2);;

let lit off sub (p, tl) lim fn =
  if lim < 0 then () else
  let db = try Lib.assoc p !db with Failure _ -> [] in
  let iter_fun (dbtl, rest, ground) =
    try
      if ground then
        let s, _ = fol_unify_l 0 tl dbtl (sub, []) in fn off s rest lim (p, tl)
      else
        let noff = Fol.offinc + off in
        let (ns, nv) = fol_unify_l noff tl dbtl (sub, []) in
        let tl = List.map (fol_subst_partial ns) tl in
        let s, rest = optimize_subst noff (ns, nv) rest in fn noff s rest (lim - 1) (p, tl)
    with Failure _ -> () in
  List.iter iter_fun db;;

let cut1 = true;;
let cut2 = true;;
let cut3 = true;;

exception Cut;;
let iter = ref 0;;
type proof = Lem of fol_lit | Pat of fol_lit | Res of fol_lit * int;;
exception Solved of proof list
(* print_fof_clause (List.map (Fol.fol_inst subst) clause); print_string "\t\t"; print_fof_clause (List.rev (List.map (Fol.fol_inst subst) path)); print_string "\t\t"; print_fol_subst subst; print_char '\n';*)

let ss s (x, l) = (x, List.map (fol_subst_partial s) l);;
let ssl s = List.map (fun (x, l) -> (x, List.map (fol_subst_partial s) l));;

let rec prove off subst path limit lemmas todo = function
  [] -> begin match todo with
      [] -> raise (Solved [])
    | None :: t -> prove off subst path limit lemmas t []; raise Cut
    | Some (p,pl,l,c) :: t -> prove off subst p pl l t c
    end
| ((lit1 :: rest_clause) as clause) ->
(* print_fof_clause (ssl subst clause); print_string "\t\t"; print_fof_clause (List.rev (ssl subst path)); print_string "\t\t"; print_fol_subst subst; print_char '\n';*)
      if (List.exists (fun x -> List.exists (fol_atom_eq subst x) path)) clause then () else

      if List.exists (fol_atom_eq subst lit1) lemmas && ((try prove off subst path limit lemmas todo rest_clause with Solved x -> raise (Solved ((Lem (ss subst lit1)) :: x))); cut1) then () else

      let neglit = mk_negated lit1 in
      if List.fold_left (fun sf plit -> if sf then true else
        try let (s, _) = fol_unify_lit 0 neglit plit subst in
        ((try prove off s path limit (lit1 :: lemmas) todo rest_clause with Solved x -> raise (Solved ((Pat (ss s lit1)) :: x))); true) with Failure _ -> sf) false path && cut2 then () else

      (try lit off subst neglit limit (fun noff s cla1 new_limit neglit1 -> let lit1 = mk_negated neglit1 in
        let todo = Some (path, limit, lit1 :: lemmas, rest_clause) :: todo in
        try prove noff s (lit1 :: path) new_limit lemmas (if cut3 then None :: todo else todo) cla1 with Solved x -> raise (Solved (Res (ss s lit1, List.length cla1) :: x))
      ) with Cut -> ());;

(* FIXME: Think about the order in 'imadd' *)
let axioms2db rev axs =
  let rec iter_rest acc fnctn = function
    [] -> ()
  | h :: t -> fnctn h (List.filter (fun x -> x <> hashf) (List.rev_append acc t)); iter_rest (h :: acc) fnctn t in
  db := [];
  let imadd k v =
    let ol, db' = try List.assoc k !db, List.remove_assoc k !db with _ -> [], !db in db := (k, if rev then v :: ol else ol @ [v]) :: db'
  in
  let cl2db cl =
    let ground = Lib.forall (fun (_, t) -> Lib.forall fol_ground t) cl in
    iter_rest [] (fun (p,tl) rest -> imadd p (tl, rest, ground)) cl in
  List.iter cl2db axs;;


let file_to_problem fname =
  let fofs = Fof_lexer.read_fof fname in
  let _, ths = List.split (List.map (fun (_,n,_,x) -> n, x) (List.filter (fun (_, _, x, _) -> x = "axiom") fofs)) in
  let gl = try Bool.list_mk_disj (List.map (fun (_,_,_,x) -> x) (List.filter (fun (_, _, x, _) -> x = "conjecture") fofs)) with _ -> failwith "No conjecture!" in
  (ths, gl);;

let prob_to_lean_dnf (ths, gl) =
  let ths = ths @ List.map Fusion.concl (Fol.create_equality_axioms (gl :: ths)) in
  let skolem1 tm = snd (Bool.dest_iff (Fusion.concl ((Equal.tHENC (Equal.tHENC Canon.nNF_CONV Canon.sKOLEM_CONV) Canon.pRENEX_CONV) tm))) in
  let skolem2 tm = snd (Basics.strip_exists (skolem1 tm)) in
  let cnf tm = snd (Bool.dest_iff (Fusion.concl (Canon.cNF_CONV (skolem2 tm)))) in
  let dnf tm = snd (Bool.dest_iff (Fusion.concl (Canon.nNF_CONV (Bool.mk_neg (cnf (Bool.mk_neg tm)))))) in
  let conj = dnf (Bool.mk_conj (hash, gl)) in
  let axs = dnf (Bool.mk_neg (Bool.list_mk_conj (hash :: ths))) in
  let (cv, conj) = Basics.strip_exists conj
  and (av, axs) = Basics.strip_exists axs in
  let cs = List.map (fun d -> List.map (Fol.fol_of_literal true [] cv) (Basics.conjuncts d)) (Basics.disjuncts conj) in
  let ds = List.map (fun d -> List.map (Fol.fol_of_literal true [] av) (Basics.conjuncts d)) (Basics.disjuncts axs) in
  List.filter (fun x -> x <> [hashfn]) (List.rev_append cs ds);;

exception Alarm;;
let alarm _ = Printf.printf ".%!"; raise Alarm;;
Sys.signal Sys.sigalrm (Sys.Signal_handle alarm);;
let maxsec i f a = ignore (Unix.alarm i); try f a; ignore (Unix.alarm 0); () with Alarm -> ();;

let randdb () =
  let v (t,_,g) = if g then Random.int 10000 - 10000 else if Lib.forall fol_ground t then Random.int 10000 else Random.int 10000 + 10000 in
  db := List.map (fun (x,y) -> (x,List.sort (fun a b -> compare (v a) (v b)) y)) !db
;;

let leancop_problem prob =
  let axioms = prob_to_lean_dnf prob in
  List.iter (fun x -> print_fof_clause x; print_char '\n') axioms;
  Printf.printf "----------------\n%!";
  axioms2db false axioms;
  maxsec 1000 (fun () -> for i = 1 to 1000 do print_char '+'; prove 10000 Im.empty [] i [] [] [hashfn] done) ();
  (*randdb ();
  maxsec 10 (fun () -> for i = 1 to 1000 do prove 10000 Im.empty [] i [] [] [hashfn] done) ();
  randdb ();
  maxsec 10 (fun () -> for i = 1 to 1000 do prove 10000 Im.empty [] i [] [] [hashfn] done) ();*)
  failwith "over 1000"
;;

(*let meson_problem (ths, gl) =
  Meson.mESON (List.map (fun x -> Fusion.Sequent ([], x)) ths) gl
;;*)

try leancop_problem (file_to_problem (Sys.argv.(1)))
with Solved pl -> Printf.printf "\nSZS status Theorem\n%!";
  (let print_step = function
    | Lem x -> print_string "Lem: "; print_fof_clause [x]; print_char '\n'
    | Pat x -> print_string "Pat: "; print_fof_clause [x]; print_char '\n'
    | Res (x,i) -> print_string "Res: "; print_fof_clause [x]; print_char '\n' in
  List.iter print_step pl)
  | Failure x -> Printf.printf "\nSZS status Unknown\n%s\n%!" x
;;

(*meson_problem (file_to_problem (Sys.argv.(1)));;*)





