open Ncmatrix
open Utils

type 'v litmat_ext =
    (* beta clause (left & right), whole clause *)
    Litext of ('v clause * 'v clause) * 'v clause
    (* matrix index, surrounding clause, path in clause *)
  | Matext of int * ('v clause * 'v clause) * 'v clause_ext
(* clause index, surrounding matrix, path in litmat *)
and 'v clause_ext = (int * 'v list) * ('v matrix * 'v matrix) * 'v litmat_ext

let offset_iv off (i, v) = (i, List.map ((+) off) v)

let rec offset_litmat_ext off = function
    Litext (claB, claC) ->
      Litext (Pair.mapn (offset_clause off) claB, offset_clause off claC)
  | Matext (j, claLR, clext) ->
      Matext (j, Pair.mapn (offset_clause off) claLR, offset_clause_ext off clext)
and offset_clause_ext off (iv, matLR, lmext) =
  offset_iv off iv, Pair.mapn (offset_matrix off) matLR, offset_litmat_ext off lmext

let rec claBC_of_litmat_ext = function
    Litext ((claL, claR), claC) -> (List.rev_append claL claR, claC)
  | Matext (j, (claL, claR), clext) ->
      let matB, matC = matBC_of_clause_ext clext
      and claLR = List.rev_append claL claR in
      (if !Arglean.Nano.middle_test then List.rev_append claL (Mat (j, matB) :: claR) else
       if !Arglean.Nano.end_test then List.rev_append claL (claR @ [Mat (j, matB)]) else
      Mat (j, matB) :: claLR),
      Mat (j, matC) :: claLR
and matBC_of_clause_ext (iv, (matL, matR), lmext) =
  let claB, claC = claBC_of_litmat_ext lmext in
  [iv, claB], (iv, claC) :: List.rev_append matL matR

let rec litext_of_litmat_ext = function
    Litext (claB, claC) -> claB, claC
  | Matext (_, _, clext) -> litext_of_clause_ext clext
and litext_of_clause_ext (_, _, lmext) = litext_of_litmat_ext lmext

let rec assert_iclause matLR (iv, (c : 'a clause)) =
  if !Arglean.Trace.verbose then Format.printf "Asserting clause %s.\n%!" (Ncprint.string_of_nclause c);
  List.list_rest c |> List.concat_map (fun (lm, claLR) ->
    map_litmat
      (fun lit -> [lit, Litext (claLR, c)])
      (assert_imatrix claLR) lm)
    |> List.map (fun (lit, lmext) -> lit, (iv, matLR, lmext))
and assert_imatrix claLR (j, mat) =
  List.list_rest mat |> List.concat_map (fun (cla, matLR) -> assert_iclause matLR cla)
    |> List.map (fun (lit, claext) -> lit, Matext (j, claLR, claext))

let assert_matrix = List.concat_map (assert_iclause ([], []))

