open Fof
open Term
open Utils


(* Expects nnf with no quantifiers *)
let rec cnf = function
    Disj (Conj (a, b), c) -> Conj (cnf (Disj (a, c)), cnf (Disj (b, c)))
  | Disj (a, Conj (b, c)) -> Conj (cnf (Disj (a, b)), cnf (Disj (a, c)))
  | Disj (a, b) -> (match cnf a, cnf b with
        (Conj _ as a1), b1 | a1, (Conj _ as b1) -> cnf (Disj (a1, b1))
      | a1, b1 -> Disj (a1, b1))
  | Conj (a, b) -> Conj (cnf a, cnf b)
  | (Atom _ as x) | (Neg (Atom _) as x) -> x
  | _ -> failwith "cnf"

(*$Q
  Fof_random.arbitrary_form (fun t -> let t' = Fof.nnf t in cnf t' = cnf (cnf t'))
*)

let rec merge_def_nnf (def, frm) = match def with
    [] -> frm
  | (a, (Conj (b, c))) :: tl ->
      merge_def_nnf ((a, b) :: (a, c) :: tl, frm)
  | (a, b) :: tl -> merge_def_nnf (tl, Conj (Disj (a, b), frm))

let def_nnf paths dname =
  let maybe_intro_def fv ((def, nnf), p) = match (nnf, dname) with
      Conj (_, _), Some dname ->
        let atom = Atom (dname nnf, List.map (fun v -> V v) fv) in
        ((Neg atom, nnf) :: def, atom), p
    | _, _ -> (def, nnf), p in

  let rec aux fv = let open Big_int in function
      Conj (a, b) ->
        let ((defa, a'), pa) = aux fv a
        and ((defb, b'), pb) = aux fv b in
        let nnf = if paths && gt_big_int pa pb then Conj (b', a') else Conj (a', b') in
        ((defa @ defb, nnf), mult_big_int pa pb)
    | Disj (a, b) ->
        let ((defa, a'), pa) = maybe_intro_def fv (aux fv a)
        and ((defb, b'), pb) = maybe_intro_def fv (aux fv b) in
        let nnf = if paths && gt_big_int pa pb then Disj (b', a') else Disj (a', b') in
        ((defa @ defb, nnf), add_big_int pa pb)
    | Forall (x, t) -> aux (x :: fv) t
    | (Atom _ as x) | (Neg (Atom _) as x) -> (([], x), unit_big_int)
    | _ -> failwith "def_nnf"

  in aux [] %> fst %> merge_def_nnf


let triv_clause cl =
  List.exists (fun (p1,a1) -> List.exists (fun (p2,a2) -> p1 = -p2 && a1 = a2) cl) cl

let rec mat_of_cnf = function
    Conj (a, b) -> mat_of_cnf a @ mat_of_cnf b
  | Disj (a, b) -> begin match (mat_of_cnf a, mat_of_cnf b) with
      ([ca], [cb]) -> let c = List.union2 ca cb in if triv_clause c then [] else [c]
    | _ -> [] end
  | Neg (Atom (p, a)) -> [[-p, a]]
  | Atom t -> [[t]]
  | _ -> failwith "mat_of_cnf"


let rename_clause l = snd (List.fold_map rename_lit (IM.empty, 0) l)


let clause_positive cl = List.for_all (fun (p, _) -> p < 0) cl

let clause_max_var cl = 1 + List.fold_left lit_max_var (-1) cl
