open Print
open Substitution.Substlist
open Format
open Utils

type ('l, 'c, 're) proof =
    Lemma
  | Reduction
  | Decomposition of 'c *       ('l * ('l, 'c, 're) proof) list
  | Extension     of 'c * 're * ('l * ('l, 'c, 're) proof) list

let rec pp_print_proof (fl, fc) sub f (lit, prf) =
  pp_open_vbox f 2;
  pp_print_string f (fl sub lit); pp_print_char f ' ';
  pp_print_prf (fl, fc) sub f prf;
  pp_close_box f ()
and pp_print_prf (fl, fc) sub f = function
    Reduction -> pp_print_string f "Red"
  | Lemma -> pp_print_string f "Lem"
  | Decomposition (cl, prfs) ->
      pp_print_string f ("Dec " ^ fc cl);
      pp_print_proofs (fl, fc) sub f prfs
  | Extension (cl, recon, prfs) ->
      pp_print_string f ("Ext " ^ fc cl);
      pp_print_proofs (fl, fc) sub f prfs
and pp_print_proofs (fl, fc) sub f =
  List.iter (fun x -> pp_print_cut f (); pp_print_proof (fl, fc) sub f x)

let print_proof (fl, fc) sub prf = pp_nl (pp_print_proof (fl, fc) sub) std_formatter prf

let string_of_sub_lit sub lit = string_of_lit (inst_lit sub lit)
let string_of_sub_litmat sub =
  Ncmatrix.map_litmat (string_of_sub_lit sub) (snd %> Ncprint.string_of_matrix)

let print_proof_ncl prf =
  print_proof (string_of_sub_litmat, Ncprint.string_of_nclause_ic) prf
