open Mapping
open Fof
open Utils

let th_types =
  ["axiom"; "hypothesis"; "definition"; "lemma"; "theorem"; "plain"]

let forms_with_type p ths =
  List.map (fun (_, _, x) -> x) (List.filter (fun (_, x, _) -> p x) ths)

let conjoin_forms = function
    [] -> None
  | l -> Some (List.fold_right1 conj l)

let split_goal fofs =
  let ths = forms_with_type (fun x -> List.mem x th_types) fofs
  and cjs = forms_with_type ((=) "conjecture") fofs
  and ncs = forms_with_type ((=) "negated_conjecture") fofs in
  let th = conjoin_forms ths
  and gl = conjoin_forms (List.map neg ncs @ cjs) in
  (th, gl)

let combine_goal = function
    (Some th, Some gl) -> Impl (th, gl)
  | (Some th, None) -> Neg th
  | (None, Some gl) -> gl
  | (None, None) -> failwith "combine_goal"

let sform_of_file = Fof_lexer.file %> List.rev %> split_goal %> combine_goal
