open Cnf;;
open Fof_parse;;
open Fof;;
open Features;;
open Proof.Clausal.Step
open Term
open Database.Clausal.ListDb
open Mapping
open Common
open Arglean.Cop
open Utils

module Stacksearch : Common.Search =
struct

type result = Substitution.Substlist.t * Proof.Clausal.Step.proof_step list
let process = Option.map Proof.Clausal.Proof.of_sub_steps %> show_result
let load_db sform (conj, def) = axioms2db (mat_of_sform conj def sform)

module Make (Sub : Substitution.Substitution) =
struct

module Subst = Substitution.Substoff (Sub)

exception Cut of int

let catch_cut n f arg =
  try f arg with Cut m -> if n = m then () else raise (Cut m)

exception Solved of Subst.t * proof_step list

let rec prove (sub, off) path lim lem stack = function
  [] -> begin match stack with
      [] -> raise (Solved (sub, []))
    | (cut, p, lim, l, c) :: t ->
        prove (sub, off) p lim l t c; if !cut3 then raise (Cut cut)
    end
| (lit1 :: cla) as clause ->

    if (List.exists (fun x -> List.exists (Subst.eq_lit sub x) path)) clause then () else

    if List.exists (Subst.eq_lit sub lit1) lem &&
       ((try prove (sub, off) path lim lem stack cla
         with Solved (s,x) -> raise (Solved (s, Lem lit1 :: x)));
        !cut1) then () else

    let neglit = negate_lit lit1 in

    if List.fold_left (fun sf plit -> if sf then true else
      try let sub = Subst.unify_lit sub neglit plit in
      ((try prove (sub, off) path lim (lit1 :: lem) stack cla
        with Solved (s,x) -> raise (Solved (s, Pat lit1 :: x)));
       !cut2)
      with Subst.Unify -> sf) false path then () else

    let db = db_entries neglit in

    let iter_fun (lit2, cla2, vars, hsh) =
      if lim >= 0 || vars = 0 then
        (match Subst.unify_rename (sub, off) (snd lit1) (lit2, cla2, vars, hsh) with
          Some (s, cla1) ->
            incr Stats.infer;
            let stack = ((if !cut3 then lim else -1), path, lim, lit1 :: lem, cla) :: stack in
            (try prove s (lit1 :: path) (lim - 1) lem stack cla1
             with Solved (s,x) -> raise (Solved (s, Res (lit1, path, lem, hsh) :: x)))
        | None -> ()) in
    if !cut3 then catch_cut lim (List.iter iter_fun) db
    else List.iter iter_fun db

let start lim =
  try prove (Subst.empty 1000000, 0) [] lim [] [] [(hash, [])]; None
  with Solved (sub, prf) -> Some (Subst.to_list sub, List.rev prf)

end
end

module Stackcop = Common.Cop (Stacksearch)

let _ =
  let speclist = Arg.align Arglean.(Prematrix.args @ Lean.args @ Cop.args) in
  let usage = "Usage: stackcop [options] <file.p>\nAvailable options are:" in
  Common.main speclist usage Stackcop.search

(*
let run file =
  let (module Searchimpl : Search) = if true then (module Stacksearch) else (module Stacksearch) in
  let module Run = Cop(Searchimpl) in
  Run.search file
*)

