type 'a strom = unit -> 'a
exception End_of_strom

let nil () = raise End_of_strom

let cons x l =
  let past = ref false in
  (fun () -> if !past then l () else (past := true; x))

let append xs ys =
  let past = ref false in
  (fun () -> if !past then ys ()
             else try xs () with End_of_strom -> (past := true; ys ()))

let of_list l =
  let r = ref l in
  (fun () -> match !r with
      [] -> raise End_of_strom
    | x :: xs -> r := xs; x)

let rec to_list l = try let x = l () in x :: to_list l with End_of_strom -> []

let iter f l =
  let rec iter () = let x = l () in f x; iter () in
  try iter () with End_of_strom -> ()

let map (f : 'a -> 'b) (l : 'a strom) () = f (l ())

let mapi (f : int -> 'a -> 'b) (l : 'a strom) =
  let i = ref 0 in
  fun () -> let h = !i in incr i; f h (l ())

let take n l =
  let i = ref n in
  fun () -> if !i = 0 then raise End_of_strom else (decr i; l ())

let take_while p l () =
  let x = l () in
  if p x then x else raise End_of_strom

let rec filter p l () =
  let x = l () in
  if p x then x else filter p l ()

let rec filter_map f l () =
  match f (l ()) with
    None -> filter_map f l ()
  | Some y -> y

let concat (ls : 'a strom strom) =
  let last = ref nil in
  let rec concat () =
    try !last () with End_of_strom -> (last := ls (); concat ()) in
  concat

let rec concat_map (f : 'a -> 'b strom) (l : 'a strom) =
  let last = ref nil in
  let rec concat_map () =
    try !last () with End_of_strom -> (last := f (l ()); concat_map ()) in
  concat_map

let next_opt l = try Some (l ()) with End_of_strom -> None

let cache (f : unit -> 'a strom) =
  let result = ref None in
  fun () -> match !result with None -> let l = f () in result := Some l; l () | Some l -> l ()

let cut xs ys =
  let cut = ref false
  and past = ref false in
  fun () ->
    if !past then ys ()
    else if !cut then nil ()
    else
      try let x = xs () in (cut := true; x)
      with End_of_strom -> (past := true; ys ())

let integers_from n =
  let i = ref n in
  fun () -> let h = !i in incr i; h

(*
let _ =
  (*let s = append (of_list [1; 2]) (of_list [3; 4]) in*)
  (*iter (Format.printf "%d\n%!") (take 200000 (integers_from 0));*)

  concat_map (fun x -> Format.printf "x = %d\n%!" x; take x (integers_from 0)) (take 10000 (integers_from 0)) |>
(*
  concat (of_list [of_list [1;2]]) |>
*)
  iter (Format.printf "%d\n%!")
*)
